@@ -16,7 +16,10 @@
 # under the License.
 
 package Search::Elasticsearch::Role::Cxn;
-$Search::Elasticsearch::Role::Cxn::VERSION = '7.30';
+$Search::Elasticsearch::Role::Cxn::VERSION = '8.12';
+our $PRODUCT_CHECK_HEADER = 'x-elastic-product';
+our $PRODUCT_CHECK_VALUE = 'Elasticsearch';
+
 use Moo::Role;
 use Search::Elasticsearch::Util qw(parse_params throw to_list);
 use List::Util qw(min);
@@ -32,6 +35,7 @@ use Net::IP;
 
 requires qw(perform_request error_from_text handle);
 
+has 'client_version'        => ( is => 'ro', required => 1, default => $Search::Elasticsearch::VERSION );
 has 'host'                  => ( is => 'ro', required => 1 );
 has 'port'                  => ( is => 'ro', required => 1 );
 has 'uri'                   => ( is => 'ro', required => 1 );
@@ -81,6 +85,13 @@ sub get_user_agent {
     return sprintf("elasticsearch-perl/%s (%s; perl %s)", $Search::Elasticsearch::VERSION, $^O, $]);
 }
 
+#===================================
+sub get_meta_header {
+#===================================
+    return sprintf("es=%s,pl=%s", $Search::Elasticsearch::VERSION, $^V);
+}
+
+
 #===================================
 sub BUILDARGS {
 #===================================
@@ -138,6 +149,32 @@ sub BUILDARGS {
 
     $default_headers{'User-Agent'} = $class->get_user_agent();
 
+    # Add Elastic meta header
+    $default_headers{'x-elastic-client-meta'} = $class->get_meta_header();
+
+    # Compatibility header
+    if (defined $ENV{ELASTIC_CLIENT_APIVERSIONING} &&
+        (lc($ENV{ELASTIC_CLIENT_APIVERSIONING}) eq 'true' || $ENV{ELASTIC_CLIENT_APIVERSIONING} eq '1')) {
+            $default_headers{'Accept'} = 'application/vnd.elasticsearch+json;compatible-with=7';
+            $default_headers{'Content-Type'} = 'application/vnd.elasticsearch+json; compatible-with=7';
+    }
+
+    if (defined $params->{elastic_cloud_api_key} && defined $params->{token_api}) {
+        throw( 'Request',
+            "You cannot set elastic_cloud_api_key and token_api together" );
+    }
+
+    # Elastic cloud API key
+    if (defined $params->{elastic_cloud_api_key}) {
+        $default_headers{'Authorization'} = sprintf("ApiKey %s", $params->{elastic_cloud_api_key});
+    }
+
+    # Elasticsearch token API (https://www.elastic.co/guide/en/elasticsearch/reference/current/security-api-get-token.html)
+    if (defined $params->{token_api}) {
+        $default_headers{'Authorization'} = sprintf("Bearer %s", $params->{token_api});
+    }
+
+    # Elasticsearch
     $params->{scheme}   = $scheme;
     $params->{is_https} = $scheme eq 'https';
     $params->{host}     = $host;
@@ -147,6 +184,10 @@ sub BUILDARGS {
     $host = "[$host]" if Net::IP::ip_is_ipv6($host);
     $params->{uri}             = URI->new("$scheme://$host:$port$path");
     $params->{default_headers} = \%default_headers;
+    # Add the client version
+    if (defined $params->{client}) {
+        $params->{client_version} = substr($params->{client}, 0, index($params->{client}, '_'));
+    }
 
     return $params;
 }
@@ -323,6 +364,15 @@ sub _decompress_body {
 sub process_response {
 #===================================
     my ( $self, $params, $code, $msg, $body, $headers ) = @_;
+
+    # Product check only for 8+ client API version
+    if ( $self->client_version >= 8 and $code >= 200 and $code < 300 ) {
+        my $product = $headers->{$PRODUCT_CHECK_HEADER} // '';
+        if ($product ne $PRODUCT_CHECK_VALUE) {
+            throw( "ProductCheck", "The client noticed that the server is not Elasticsearch and we do not support this unknown product" );
+        }
+    }
+
     $self->_decompress_body( \$body, $headers );
 
     my ($mime_type) = split /\s*;\s*/, ( $headers->{'content-type'} || '' );
@@ -338,7 +388,6 @@ sub process_response {
     }
 
     # Request is successful
-
     if ( $code >= 200 and $code <= 209 ) {
         if ( defined $body and length $body ) {
             $body = $self->serializer->decode($body)
@@ -446,7 +495,7 @@ Search::Elasticsearch::Role::Cxn - Provides common functionality to HTTP Cxn imp
 
 =head1 VERSION
 
-version 7.30
+version 8.12
 
 =head1 DESCRIPTION
 
@@ -814,7 +863,7 @@ Enrico Zimuel <enrico.zimuel@elastic.co>
 
 =head1 COPYRIGHT AND LICENSE
 
-This software is Copyright (c) 2020 by Elasticsearch BV.
+This software is Copyright (c) 2024 by Elasticsearch BV.
 
 This is free software, licensed under:
 
